/* This file is part of fileserv.
   Copyright (C) 2017 Sergey Poznyakoff

   Fileserv is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   Fileserv is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with fileserv.  If not, see <http://www.gnu.org/licenses/>. */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#define MHD_PLATFORM_H
#include <microhttpd.h>

#include "fileserv.h"

char *progname;
int verbose; /* reserved for future use */
char *address = "0.0.0.0";
char *forwarded_header = "X-Forwarded-For";

#ifndef DEFAULT_SERVICE
# define DEFAULT_SERVICE "8080"
#endif

static void
fileserv_panic(void *cls, const char *file, unsigned int line,
	     const char *reason)
{
        if (reason)
		error("%s:%d: MHD PANIC: %s", file, line, reason);
	else
		error("%s:%d: MHD PANIC", file, line);
	abort();
}

void
xmalloc_fail(void)
{
	error("not enough memory");
	exit(1);
}

void *
xmalloc(size_t s)
{
	void *p = malloc(s);
	if (!p)
		xmalloc_fail();
	return p;
}

void *
xcalloc(size_t nmemb, size_t size)
{
	char *ptr;
	if (size < SIZE_T_MAX / nmemb)
		ptr = calloc(nmemb, size);
	else
		ptr = NULL;
	if (!ptr)
		xmalloc_fail();
	return ptr;
}

void *
xrealloc(void *ptr, size_t size)
{
	ptr = realloc(ptr, size);
	if (!ptr)
		xmalloc_fail();
	return ptr;
}

void
usage(void)
{
	printf("usage: %s [OPTIONS] [HOST:]URL:DIR...\n", progname);
	printf("OPTIONS are:\n\n");
	printf("  -a ADDRESS:PORT      listen on the given address\n");
	printf("  -F FACILITY          use this syslog facility\n");
	printf("  -f                   remain in the foreground\n");
	printf("  -g GROUP             run with privileges of GROUP\n");
	printf("  -h                   show this help summary\n");
	printf("  -p FILE              store PID in FILE\n");
	printf("  -t IP                add trusted IP address\n");
	printf("  -x HEADER            use HEADER instead of X-Forwarded-For\n");
	printf("  -u USER              run with this USER privileges\n");
}

int
open_node(char const *node, char const *serv, struct sockaddr **saddr)
{
	struct addrinfo hints;
	struct addrinfo *res;
	int reuse;
	int fd;
	int rc;
	
	memset(&hints, 0, sizeof (hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	rc = getaddrinfo(node, serv, &hints, &res);
	if (rc) {
		error("%s: %s", node, gai_strerror(rc));
		exit(1);
	}

	fd = socket(res->ai_family, res->ai_socktype, 0);
	if (fd == -1) {
		error("socket: %s", strerror(errno));
		exit(1);
	}

	reuse = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		   &reuse, sizeof(reuse));
	if (bind(fd, res->ai_addr, res->ai_addrlen) == -1) {
		error("bind: %s", strerror(errno));
		exit(1);
	}

	if (listen(fd, 8) == -1) {
		error("listen: %s", strerror(errno));
		exit(1);
	}

        if (saddr) {
		*saddr = xmalloc(res->ai_addrlen);
		memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	freeaddrinfo(res);

	return fd;
}

int
open_listener(char const *addr, struct sockaddr **saddr)
{
	size_t len;
	int fd;
	
	len = strcspn(addr, ":");
	if (len == 0)
		fd = open_node(NULL, addr + 1, saddr);
	else if (addr[len] == 0)
		fd = open_node(addr, DEFAULT_SERVICE, saddr);
	else {
		char *node;
		node = xmalloc(len + 1);
		memcpy(node, addr, len);
		node[len] = 0;
		fd = open_node(node, addr + len + 1, saddr);
		free(node);
	}
	return fd;
}

struct urimap {
	char *host;
	char *uri;
	size_t uri_len;
	char *dir;
	size_t dir_len;
	struct urimap *next;
};

struct urimap *map_head, *map_tail;

struct urimap const *
urimap_find(char const *host, char const *url)
{
	struct urimap *map;
	size_t len = strlen(url);
	
	for (map = map_head; map; map = map->next) {
		if (map->host && strcasecmp(map->host, host))
			continue;
		if (map->uri_len > len)
			continue;
		if (memcmp(url, map->uri, map->uri_len) == 0
		    && url[map->uri_len] == '/')
			return map;
	}
	return NULL;
}	

struct urimap const *
urimap_find_dir(char const *host, char const *filename)
{
	struct urimap *map;
	size_t len = strlen(filename);
	
	for (map = map_head; map; map = map->next) {
		if (map->host && strcasecmp(map->host, host))
			continue;
		if (map->dir_len > len)
			continue;
		if (memcmp(filename, map->dir, map->dir_len) == 0
		    && filename[map->dir_len] == '/')
			return map;
	}
	return NULL;
}	

void
urimap_add(char *arg)
{
	struct urimap *map;
	char *copy, *p;
	
	map = xmalloc(sizeof(*map) + strlen(arg) + 1);
	copy = (char*)(map + 1);
	strcpy(copy, arg);
	p = strchr(copy, ':');
	if (p) {
		*p++ = 0;
		map->dir = strchr(p, ':');
		if (map->dir) {
			map->host = copy;
			map->uri = p;
			*map->dir++ = 0;
		} else {
			map->host = NULL;
			map->uri = copy;
			map->dir = p;
		}
	} else {
		map->host = NULL;
		map->uri = copy;
		map->dir = NULL;
	}
	map->uri_len = strlen(map->uri);
	while (map->uri_len > 0 && map->uri[map->uri_len-1] == '/')
		map->uri[--map->uri_len] = 0;
	if (map->dir) {
		map->dir_len = strlen(map->dir);
		while (map->dir_len > 0 && map->dir[map->dir_len-1] == '/')
			map->dir[--map->dir_len] = 0;
	}

	map->next = NULL;
	if (map_tail)
		map_tail->next = map;
	else
		map_head = map;
	map_tail = map;
}

static char *
cfname(char const *fname)
{
	/* FIXME */
	return realpath(fname, NULL);
}

char *
get_file_name(char const *host, char const *url)
{
	struct urimap const *map;
	char *file_name;
	size_t len;
	char *cf;

	map = urimap_find(host, url);
	if (!map)
		return NULL;

	len = strlen(url) - map->uri_len + map->dir_len;
	file_name = xmalloc(len + 1);
	if (map->dir_len)
		memcpy(file_name, map->dir, map->dir_len);
	strcpy(file_name + map->dir_len, url + map->uri_len);

	cf = cfname(file_name);
	if (!cf) {
		free(cf);
		return NULL;
	} else if (strcmp(cf, file_name) == 0) {
		free(cf);
	} else if (urimap_find_dir(host, cf)) {
		free(file_name);
		file_name = cf;
	} else {
		free(file_name);
		free(cf);
		return NULL;
	}
	return file_name;
}

/* Returns 1 if ADDR is a valid string representation of IPv4 address */
static int
str_is_ipv4(const char *addr)
{
	int dot_count = 0;
	int digit_count = 0;
	
	for (; *addr; addr++) {
		if (!isascii(*addr))
			return 0;
		if (*addr == '.') {
			if (++dot_count > 3)
				break;
			digit_count = 0;
		}
		else if (!(isdigit(*addr) && ++digit_count <= 3))
			return 0;
	}
	return (dot_count == 3);
}

/* Returns 1 if ADDR is a valid IPv6 address */
static int
str_is_ipv6(const char *addr)
{
	int col_count = 0; /* Number of colons */
	int dcol = 0;      /* Did we encounter a double-colon? */
	int dig_count = 0; /* Number of digits in the last group */

	for (; *addr; addr++) {
		if (!isascii (*addr))
			return 0;
		else if (isxdigit(*addr)) {
			if (++dig_count > 4)
				return 0;
		} else if (*addr == ':') {
			if (col_count && dig_count == 0 && ++dcol > 1)
				return 0;
			if (++col_count > 7)
				return 0;
			dig_count = 0;
		}
		else
			return 0;
	}
	return (col_count == 7 || dcol);
}

/* Max. number of bytes in an IPv6 address */
#define MU_INADDR_BYTES 16

/* CIDR representation */
struct cidr {
	int family;        /* Address family */
	int len;           /* Number of bytes in the address */
	unsigned char address[MU_INADDR_BYTES];
	unsigned char netmask[MU_INADDR_BYTES];
	struct cidr *next; /* Nest CIDR in list */
};

static void
uint32_to_bytes(unsigned char *bytes, uint32_t u)
{
  int i;
  
  for (i = 0; i < 4; i++)
    {
      bytes[i] = u & 0xff;
      u >>= 8;
    }
}

static int
inaddr_to_bytes(int af, void *buf, unsigned char *bytes)
{
	uint32_t u;
  
	switch (af) {
	case AF_INET:
		memcpy(&u, buf, sizeof u);
		uint32_to_bytes (bytes, u);
		return 4;

	case AF_INET6:
		memcpy(bytes, buf, 16);
		return 16;
	}
	
	abort();
}

/* Fill in the BUF (LEN bytes long) with a network mask corresponding to
   the mask length MASKLEN */
static void
masklen_to_netmask(unsigned char *buf, size_t len, size_t masklen)
{
	int i, cnt;

	cnt = masklen / 8;
	for (i = 0; i < cnt; i++)
		buf[i] = 0xff;
	if (i == len)
		return;
	cnt = 8 - masklen % 8;
	buf[i++] = (0xff >> cnt) << cnt;
	for (; i < len; i++)
		buf[i] = 0;
}

/* Convert string STR to CIDR. Return 0 on success and -1 on failure. */
static int
str_to_cidr(char const *str, struct cidr *cidr)
{
	int rc;
	char *ipbuf;
	union {
		struct in_addr in;
		struct in6_addr in6;
	} inaddr;
	size_t len;
	char *p;
	
	p = strrchr(str, '/');
	if (p)
		len = p - str;
	else
		len = strlen (str);

	ipbuf = xmalloc(len+1);
	memcpy(ipbuf, str, len);
	ipbuf[len] = 0;

	if (str_is_ipv4(ipbuf))
		cidr->family = AF_INET;
	else if (str_is_ipv6(ipbuf))
		cidr->family = AF_INET6;
	else {
		free(ipbuf);
		error("invalid CIDR: %s", str);
		return -1;
	}
	
	rc = inet_pton(cidr->family, ipbuf, &inaddr);
	free(ipbuf);

	if (rc != 1) {
		if (rc == -1) {
			error("invalid address family");
			abort();
		}
		if (rc == 0) {
			error("invalid IPv%s address: %s",
			      cidr->family == AF_INET ? "4" : "6",
			      str);
			return -1;
		}
	}

	cidr->len = inaddr_to_bytes(cidr->family, &inaddr, cidr->address);

	if (p) {
		char *end;
		unsigned long masklen;
      
		p++;

		masklen = strtoul(p, &end, 10);
		if (*end == 0)
			masklen_to_netmask(cidr->netmask, cidr->len, masklen);
		else if ((cidr->family == AF_INET && str_is_ipv4(p))
			 || (cidr->family == AF_INET6 && str_is_ipv6(ipbuf))) {
			rc = inet_pton(cidr->family, p, &inaddr);
			if (rc != 1) {
				error("bad netmask: %s", p);
				return -1;
			}
			inaddr_to_bytes(cidr->family, &inaddr, cidr->netmask);
		} else {
			error("bad CIDR (near %s)", p);
			return -1;
		}
	} else
		masklen_to_netmask(cidr->netmask, cidr->len, cidr->len * 8);
	return 0;
}

/* A is a valid CIDR, and B is a valid IP address represented as CIDR.
   Return 0 if B falls within A */
static int
cidr_match(struct cidr *a, struct cidr *b)
{
	int i;

	if (a->family != b->family)
		return 1;
	for (i = 0; i < a->len; i++) {
		if (a->address[i] != (b->address[i] & a->netmask[i]))
			return 1;
	}
	return 0;
}

struct cidr *trusted_ip_list;

static void
trusted_ip_add(char const *s)
{
	struct cidr *cidr = xmalloc(sizeof(cidr[0]));

	if (str_to_cidr(s, cidr))
		exit(1);
	cidr->next = trusted_ip_list;
	trusted_ip_list = cidr;
}

static int
is_trusted_ip(char const *ipstr)
{
	struct cidr cidr;
	struct cidr *p;
	
	if (str_to_cidr(ipstr, &cidr))
		return 0;
	for (p = trusted_ip_list; p; p = p->next) {
		if (cidr_match(p, &cidr) == 0)
			return 1;
	}
	return 0;
}	

static char *
scan_forwarded_header(char const *hdr)
{
	char const *end;
	char *buf = NULL;
	size_t bufsize = 0;
	
	end = hdr + strlen(hdr);
	while (end > hdr) {
		char const *p;
		size_t len, i, j;
		
		for (p = end - 1; p > hdr && *p != ','; p--)
			;
		len = end - p;
		if (len + 1 > bufsize) {
			char *newbuf = realloc(buf, len + 1);
			if (newbuf) {
				buf = newbuf;
				bufsize = len + 1;
			} else {
				free(buf);
				return NULL;
			}
		}

		i = 0;
		j = 0;
		if (*p == ',')
			j++;
		while (j < len && isspace(p[j]))
			j++;
		while (j < len) {
			if (isspace(p[j]))
				break;
			buf[i++] = p[j++];
		}
		buf[i] = 0;

		if (!is_trusted_ip(buf))
			return buf;

		end = p;
	}
	free(buf);		
	return 0;
}

char *
get_remote_ip(struct MHD_Connection *conn)
{
	union MHD_ConnectionInfo const *ci;
        char ipstr[NI_MAXHOST];
	char const *hdr;
	char *ret;
	
	hdr = MHD_lookup_connection_value(conn,
					  MHD_HEADER_KIND,
					  forwarded_header);
	if (hdr && (ret = scan_forwarded_header(hdr)) != NULL)
		return ret;
	
	ci = MHD_get_connection_info(conn,
				     MHD_CONNECTION_INFO_CLIENT_ADDRESS,
				     NULL);
	if (!ci)
		return NULL;

	if (getnameinfo(ci->client_addr, sizeof(struct sockaddr),
			ipstr, sizeof(ipstr), NULL, 0,
			NI_NUMERICHOST))
		return NULL;
	return strdup(ipstr);
}

void 
http_log(struct MHD_Connection *connection,
	 char const *method, char const *url,
	 int status, struct stat const *st)
{
        char *ipstr;
	char const *host, *referer, *user_agent;
	time_t t;
	struct tm *tm;
	char tbuf[30];

	ipstr = get_remote_ip(connection);

	host = MHD_lookup_connection_value(connection,
					   MHD_HEADER_KIND,
					   MHD_HTTP_HEADER_HOST);
	referer = MHD_lookup_connection_value(connection,
					      MHD_HEADER_KIND,
					      MHD_HTTP_HEADER_REFERER);
	user_agent = MHD_lookup_connection_value(connection,
						 MHD_HEADER_KIND,
						 MHD_HTTP_HEADER_USER_AGENT);

	t = time(NULL);
	tm = localtime(&t);
	strftime(tbuf, sizeof(tbuf), "[%d/%b/%Y:%H:%M:%S %z]", tm);
	
	info("%s %s - - %s \"%s %s\" %3d %lu \"%s\" \"%s\"",
	     host, ipstr, tbuf, method, url, status, st ? st->st_size : 0,
	     referer ? referer : "",
	     user_agent ? user_agent : "");
	free(ipstr);
}

static int
not_found(struct MHD_Connection *connection,
	  char const *method, char const *url)
{
	int ret;
	struct MHD_Response *response;
	static char errtext[] = "Not found";

	http_log(connection, method, url, MHD_HTTP_NOT_FOUND, NULL);
	response = MHD_create_response_from_buffer (strlen(errtext),
						    (void *) errtext,
						    MHD_RESPMEM_PERSISTENT);
	ret = MHD_queue_response(connection, MHD_HTTP_NOT_FOUND, response);
	MHD_destroy_response(response);
	return ret;
}

static int
fileserv_handler(void *cls,
		 struct MHD_Connection *conn,
		 const char *url, const char *method,
		 const char *version,
		 const char *upload_data, size_t *upload_data_size,
		 void **con_cls)
{
	static int aptr;
	char const *host = MHD_lookup_connection_value(conn,
						       MHD_HEADER_KIND,
						       MHD_HTTP_HEADER_HOST);
	char *file_name;
	struct stat st;
	struct MHD_Response *response;
	int ret;
	int fd;
	
	if (strcmp(method, MHD_HTTP_METHOD_GET) &&
	    strcmp(method, MHD_HTTP_METHOD_HEAD))
		return MHD_NO;              /* unexpected method */

	if (&aptr != *con_cls) {
		*con_cls = &aptr;
		return MHD_YES;
	}
	*con_cls = NULL;

	file_name = get_file_name(host, url);
	if (!file_name)
		return not_found(conn, method, url);

	fd = open(file_name, O_RDONLY);
	free(file_name);

	if (fd == -1) {
		return not_found(conn, method, url);
	}
	
	if (fstat(fd, &st) || !S_ISREG(st.st_mode)) {
		close(fd);
		return not_found(conn, method, url);
	}
		
	response = MHD_create_response_from_fd64(st.st_size, fd);
	if (!response) {
		close(fd);
		return MHD_NO;
	}
	ret = MHD_queue_response(conn, MHD_HTTP_OK, response);
	MHD_destroy_response(response);
	http_log(conn, method, url, MHD_HTTP_OK, &st);
	
	return ret;
}

int
main(int argc, char **argv)
{
	int c, i;
	char *p;
	int fd = -1;
	struct MHD_Daemon *mhd;
	char *user = NULL, *group = NULL;
	struct sockaddr *server_addr;
	int foreground = 0;
	sigset_t sigs;
	static int fatal_signals[] = {
		SIGHUP,
		SIGINT,
		SIGQUIT,
		SIGTERM,
		0
	};
	
	p = strrchr(argv[0], '/');
	if (p)
		progname = p + 1;
	else
		progname = argv[0];
	
	while ((c = getopt(argc, argv, "a:F:fg:hp:t:x:u:v")) != EOF) {
		switch (c) {
		case 'a':
			address = optarg;
			break;
		case 'F':
			set_log_facility(optarg);
			break;
		case 'f':
			foreground = 1;
			break;
		case 'g':
			group = optarg;
		        break;
		case 'h':
			usage();
			exit(0);
		case 'p':
			pidfile = optarg;
			break;
		case 't':
			trusted_ip_add(optarg);
			break;
		case 'x':
			forwarded_header = optarg;
			break;
		case 'u':
			user = optarg;
			break;
		case 'v':
			verbose++;
			break;
		default:
			exit(1);
		}
	}

	pidfile_check();
	
	fd = open_listener(address, &server_addr);

	runas(user, group);
	
	if (optind == argc) {
		error("no mappings");
		exit(1);
	}
	
	for (i = optind; i < argc; i++)
		urimap_add(argv[i]);

	if (!foreground) {
		int i;

		if (daemon(0, 1)) {
			error("daemon failed: %s", strerror(errno));
			exit(1);
		}
		for (i = 0; i < fd; i++)
			close(i);
		syslog_enable();
	}
	pidfile_create();
	
	/* Block the 'fatal signals' and SIGPIPE in the handling thread */
	sigemptyset(&sigs);
	for (i = 0; fatal_signals[i]; i++)
		sigaddset(&sigs, fatal_signals[i]);
	sigaddset(&sigs, SIGPIPE);
	
	pthread_sigmask(SIG_BLOCK, &sigs, NULL);
	MHD_set_panic_func(fileserv_panic, NULL);

	mhd = MHD_start_daemon(MHD_USE_INTERNAL_POLLING_THREAD
			       | MHD_USE_ERROR_LOG, 0,
			       fileserv_acl, server_addr, 
			       fileserv_handler, NULL,
			       MHD_OPTION_LISTEN_SOCKET, fd,
			       MHD_OPTION_EXTERNAL_LOGGER, fileserv_logger, NULL,
			       MHD_OPTION_END);
	/* Unblock only the fatal signals */
	sigdelset(&sigs, SIGPIPE);
	pthread_sigmask(SIG_UNBLOCK, &sigs, NULL);
	/* Wait for signal to arrive */
	sigwait(&sigs, &c);
	MHD_stop_daemon(mhd);
	pidfile_remove();
	return 0;
}

